﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Microsoft.Kinect;
using Microsoft.Speech.Recognition;
using System.Threading;
using System.IO;
using Microsoft.Speech.AudioFormat; 
using System.Diagnostics;
using System.Windows.Threading;
using Microsoft.Samples.Kinect.SwipeGestureRecognizer; //腕でスライド変化のためのdll
using Geis.Win32; //マウス操作のためのdll
using System.Drawing;
//using Microsoft.Kinect.Toolkit;

namespace PowerPointSlideShow
{
    public partial class SlideShow : Window //日本語 ver
    {

        /// スライド変化の識別                 
        private readonly Recognizer activeRecognizer;

        //Kinectセンサーからの画像処理を受け取るバッファ
        private byte[] pixelBuffer = null;

        //画像表示するビットマップ
        private WriteableBitmap bmpBuffer = null;

        ///Kinectセンサースケルトン取得
        private Skeleton[] skeletons = new Skeleton[0];

        //Kinectセンサー取得
        KinectSensor k;

        //スケルトン取得判定のためのフラグ
        Boolean skeletonFlag = false;
        
        //音声認識
        SpeechRecognitionEngine speechRecognizer;
        DispatcherTimer readyTimer;

        readonly int Bgr32BytesPerPixel = PixelFormats.Bgr32.BitsPerPixel / 8;

        public SlideShow()
        {
            InitializeComponent();

            // プレゼンの腕の動き認識            
            this.activeRecognizer = this.CreateRecognizer();

            //音声認識
            InitializeSpeechRecognition();
        }

        //起動確認
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            k = KinectSensor.KinectSensors.FirstOrDefault();
            try
            {
                //Kinectの接続台数が0だったら例外を投げる
                if (KinectSensor.KinectSensors.Count == 0)
                    throw new Exception("Kinectが接続されていません!!\nKinectを接続してください!!");

                //カラーストリームの有効化
                k.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30);
                k.SkeletonStream.Enable();

                //バッファの初期化
                pixelBuffer = new byte[k.ColorStream.FramePixelDataLength];
                bmpBuffer = new WriteableBitmap(k.ColorStream.FrameWidth, k.ColorStream.FrameHeight, 96, 96, PixelFormats.Bgr32, null);
                rgbImage.Source = bmpBuffer;

                //イベントの開始
                k.SkeletonStream.Enable();
                k.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(k_ColorFrameReady);
                k.SkeletonFrameReady += new EventHandler<SkeletonFrameReadyEventArgs>(k_SkeletonFrameReady);

                k.Start();

                k.SkeletonStream.Enable(new TransformSmoothParameters()
                {
                    Smoothing = 0.7f,
                    Correction = 0.3f,
                    Prediction = 0.4f,
                    JitterRadius = 0.10f,
                    MaxDeviationRadius = 0.5f,
                });
            }
            catch (Exception ex) //例外処理
            {
                MessageBox.Show(ex.Message, "エラー", MessageBoxButton.OK, MessageBoxImage.Error);
                Close();
            }
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            
        }
        
        /// Kinectの動作を停止する
        private void StopKinect(KinectSensor kinect)
        {
            if (kinect != null)
            {
                if (kinect.IsRunning)
                {
                    k.ColorFrameReady -= new EventHandler<ColorImageFrameReadyEventArgs>(k_ColorFrameReady);
                    k.SkeletonFrameReady -= new EventHandler<SkeletonFrameReadyEventArgs>(k_SkeletonFrameReady);
                    k.SkeletonFrameReady -= new EventHandler<SkeletonFrameReadyEventArgs>(k_Mouse);

                    kinect.Stop();
                    kinect.Dispose();
                }
            }
        }

        //骨格認識の確認
        private Boolean isSkeletonTracking(Boolean flag)
        {
            if (flag)
            {
                //骨格を認識すると緑
                ellipse1.Fill = new SolidColorBrush(System.Windows.Media.Color.FromRgb(0, 255, 0));
                flag = false;
            }
            else
                //骨格を認識しないと黒
                ellipse1.Fill = new SolidColorBrush(System.Windows.Media.Color.FromRgb(0, 0, 0));
            return flag;
        }

        //スライド変化
        private Recognizer CreateRecognizer()
        {
            var recognizer = new Recognizer();

            // 進む            
            recognizer.SwipeRightDetected += (s, e) =>
            {
                System.Windows.Forms.SendKeys.SendWait("{Right}");
                ellipse1.Fill = new SolidColorBrush(System.Windows.Media.Color.FromRgb(255, 0, 0)); //色が一瞬赤に変化
            };

            // 戻る            
            recognizer.SwipeLeftDetected += (s, e) =>
            {
                System.Windows.Forms.SendKeys.SendWait("{Left}");
                ellipse1.Fill = new SolidColorBrush(System.Windows.Media.Color.FromRgb(0, 0, 255)); //色が一瞬青に変化
            };
            return recognizer;
        }
        
        //RGBカメラ表示
        void k_ColorFrameReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame imageFrame = e.OpenColorImageFrame())
            {
                if (imageFrame != null)
                {
                    int frmWidth = imageFrame.Width;
                    int frmHeight = imageFrame.Height;

                    imageFrame.CopyPixelDataTo(pixelBuffer);

                    Int32Rect src = new Int32Rect(0, 0, frmWidth, frmHeight);
                    bmpBuffer.WritePixels(src, pixelBuffer, frmWidth * 4, 0);
                }
            }
        }

        //骨格認識
        void k_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            skeletonFlag = isSkeletonTracking(skeletonFlag);
            using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
            {
                if (skeletonFrame != null)
                {
                    skeletons = new Skeleton[skeletonFrame.SkeletonArrayLength];
                    skeletonFrame.CopySkeletonDataTo(skeletons);
                }
                if (skeletons.Length != 0)
                {
                    foreach (Skeleton skel in skeletons)
                    {
                        if (skel.TrackingState == SkeletonTrackingState.Tracked)
                        {
                            skeletonFlag = true;
                            activeRecognizer.Recognize(sender, skeletonFrame, this.skeletons);
                        }
                    }
                }
            }
        }
        
        //マウス起動
        void k_Mouse(object sender, SkeletonFrameReadyEventArgs e)
        {
                using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
                {
                    if (skeletonFrame != null)
                    {
                        // トラッキングされているスケルトンのジョイントを描画する
                        Skeleton skeleton = skeletonFrame.GetFirstTrackedSkeleton();

                        if ((skeleton != null) && (skeleton.TrackingState == SkeletonTrackingState.Tracked))
                        {
                            Joint hand = skeleton.Joints[JointType.HandRight];
                            if (hand.TrackingState == JointTrackingState.Tracked)
                            {
                                // Frame中の手の位置をディスプレイの位置に対応付ける
                                ColorImagePoint point = k.MapSkeletonPointToColor(hand.Position,        // スケルトン座標 → RGB画像座標
                                  k.ColorStream.Format);
                                System.Windows.Forms.Screen screen = System.Windows.Forms.Screen.AllScreens[0];   // メインディスプレイの情報を取得
                                point.X = (point.X * screen.Bounds.Width) / k.ColorStream.FrameWidth * 2;
                                point.Y = (point.Y * screen.Bounds.Height) / k.ColorStream.FrameHeight * 2;

                                // マウスカーソルの移動
                                SendInput.MouseMove(point.X, point.Y, screen);
                            }
                        }
                    }
                }
        }

        //音声認識
        void Current_Exit(object sender, ExitEventArgs e)
        {
            if (speechRecognizer != null)
            {
                speechRecognizer.RecognizeAsyncCancel();
                speechRecognizer.RecognizeAsyncStop();
            }
            if (k != null)
            {
                k.AudioSource.Stop();
                k.Stop();
                k.Dispose();
                k = null;
            }
        }

        //指定した認識器を取得する
        private static RecognizerInfo GetKinectRecognizer()
        {
            Func<RecognizerInfo, bool> matchingFunc = r =>
            {
                string value;
                r.AdditionalInfo.TryGetValue("Kinect", out value);
                return "True".Equals(value, StringComparison.InvariantCultureIgnoreCase) && "ja-JP".Equals(r.Culture.Name, StringComparison.InvariantCultureIgnoreCase); //日本語音声認識(ja-JP)
            };
            return SpeechRecognitionEngine.InstalledRecognizers().Where(matchingFunc).FirstOrDefault();
        }

        private void InitializeSpeechRecognition()
        {
            RecognizerInfo ri = GetKinectRecognizer();
            if (ri == null)
            {
                MessageBox.Show(
                    @"Microsoft音声認識SDKがインストールされていることを確認してください",
                    "Speech SDKのロードに失敗しました",
                    MessageBoxButton.OK,
                    MessageBoxImage.Error);
                return;
            }

            try
            {
                speechRecognizer = new SpeechRecognitionEngine(ri.Id);
            }
            catch
            {
                MessageBox.Show(
                    @"Microsoft音声認識SDKがインストールされていることを確認してください",
                    "Speech SDKのロードに失敗しました",
                    MessageBoxButton.OK,
                    MessageBoxImage.Error);
            }

            //単語の登録
            var phrases = new Choices();
            phrases.Add("はじめ"); //スライドショー開始
            phrases.Add("おわり"); //スライドショー終了
            phrases.Add("つぎ"); //スライドすすむ
            phrases.Add("まえ"); //スライドもどる
            phrases.Add("くろ"); //画面黒
            phrases.Add("しろ"); //画面白
            phrases.Add("ぺん"); //ペン起動
            phrases.Add("けしごむ"); //けしごむ起動
            phrases.Add("やじるし"); //矢印起動
            phrases.Add("はき"); //インク破棄
            phrases.Add("ほじ"); //インク保持
            phrases.Add("ドラッグ"); //マウスドラッグ
            phrases.Add("ドロップ"); //マウスドロップ
            phrases.Add("もどる"); //マウス機能終了＆画面を元に戻す
            phrases.Add("マウス"); //マウス機能起動

            var gb = new GrammarBuilder();
            //Specify the culture to match the recognizer in case we are running in a different culture.
            //異なった環境の中で実行されている場合に認識エンジンを一致する環境を指定                                
            gb.Culture = ri.Culture;
            gb.Append(phrases);

            // Create the actual Grammar instance, and then load it into the speech recognizer.
            //実際の文法のインスタンスを作成し音声認識にロードする
            var g = new Grammar(gb);

            speechRecognizer.LoadGrammar(g);
            speechRecognizer.SpeechRecognized += SreSpeechRecognized;
            speechRecognizer.SpeechHypothesized += SreSpeechHypothesized;
            speechRecognizer.SpeechRecognitionRejected += SreSpeechRecognitionRejected;

            this.readyTimer = new DispatcherTimer();
            this.readyTimer.Tick += this.ReadyTimerTick;
            this.readyTimer.Interval = new TimeSpan(0, 0, 4);
            this.readyTimer.Start();
        }

        private void ReadyTimerTick(object sender, EventArgs e)
        {
            this.StartSpeechRecognition();
            this.readyTimer.Stop();
            this.readyTimer = null;
        }

        private void StartSpeechRecognition()
        {
            if (k == null || speechRecognizer == null)
                return;

            var audioSource = this.k.AudioSource;
            audioSource.BeamAngleMode = BeamAngleMode.Adaptive;
            var kinectStream = audioSource.Start();

            speechRecognizer.SetInputToAudioStream(
                    kinectStream, new SpeechAudioFormatInfo(EncodingFormat.Pcm, 16000, 16, 1, 32000, 2, null));
            speechRecognizer.RecognizeAsync(RecognizeMode.Multiple);
        }

        void SreSpeechRecognitionRejected(object sender, SpeechRecognitionRejectedEventArgs e)
        {
            Trace.WriteLine("\nSpeech Rejected, confidence: " + e.Result.Confidence);
        }

        void SreSpeechHypothesized(object sender, SpeechHypothesizedEventArgs e)
        {
            Trace.Write("\rSpeech Hypothesized: \t{0}", e.Result.Text);
        }

        void SreSpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {
            //This first release of the Kinect language pack doesn't have a reliable confidence model, so we don't use e.Result.Confidence here.
            //Kinectの言語パックの最初のリリースでは、信頼性のあるモデルを持っていないのでe.Result.Confidenceを使用してはいけない
            if (e.Result.Confidence < 0.70)
            {
                Trace.WriteLine("\nSpeech Rejected filtered, confidence: " + e.Result.Confidence);
                return;
            }

            Trace.WriteLine("\nSpeech Recognized, confidence: " + e.Result.Confidence + ": \t{0}", e.Result.Text);

            if (e.Result.Text == "はじめ") //スライドショー開始
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("{F5}");
                });
            }
            else if (e.Result.Text == "おわり") //スライドショー終了
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    //System.Windows.Forms.SendKeys.SendWait("{Esc}");
                });
            }
            else if (e.Result.Text == "つぎ") //スライドショーすすむ
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("{Right}");
                    ellipse1.Fill = new SolidColorBrush(System.Windows.Media.Color.FromRgb(204, 0, 0));
                });
            }
            else if (e.Result.Text == "まえ") //スライドショーもどる
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("{Left}");
                    ellipse1.Fill = new SolidColorBrush(System.Windows.Media.Color.FromRgb(0, 0, 255));
                });
            }
            else if (e.Result.Text == "くろ") //画面黒
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("{B}");
                });
            }
            else if (e.Result.Text == "しろ") //画面白
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("{W}");
                });
            }
            else if (e.Result.Text == "もどる") //マウス機能終了＆画面を元に戻す
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("{X}");
                    k.SkeletonFrameReady -= new EventHandler<SkeletonFrameReadyEventArgs>(k_Mouse);
                });
            }
            else if (e.Result.Text == "ぺん") //ペン起動
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("+{F10}oh");
                });
            }
            else if (e.Result.Text == "けしごむ") //けしごむ起動
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("+e");
                });
            }
            else if (e.Result.Text == "やじるし") //矢印起動
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("^a");
                });
            }
            else if (e.Result.Text == "はき") //インク破棄
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("+D");
                });
            }
            else if (e.Result.Text == "ほじ") //インク保持
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    System.Windows.Forms.SendKeys.SendWait("+K");
                });
            }
            else if (e.Result.Text == "ドラッグ") //マウスドラッグ
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    SendInput.MouseDown();
                });
            }
            else if (e.Result.Text == "ドロップ") //マウスドロップ
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    SendInput.MouseUp();
                });
            }
            else if (e.Result.Text == "マウス") //マウス機能起動
            {
                this.Dispatcher.BeginInvoke((Action)delegate
                {
                    k.SkeletonFrameReady += new EventHandler<SkeletonFrameReadyEventArgs>(k_Mouse);
                });
            }
        }
    }

    //はじめにトラッキングしたスケルトンを取得するメソッド
    public static class SkeletonExtensions
    {
        public static Skeleton GetFirstTrackedSkeleton(this SkeletonFrame skeletonFrame)
        {
            Skeleton[] skeleton = new Skeleton[skeletonFrame.SkeletonArrayLength];
            skeletonFrame.CopySkeletonDataTo(skeleton);
            return (from s in skeleton
                    where s.TrackingState == SkeletonTrackingState.Tracked
                    select s).FirstOrDefault();
        }
    }
}

